@@ -22,9 +22,9 @@
 #include <exception>
 #include <fcitx-config/rawconfig.h>
 #include <fcitx-utils/capabilityflags.h>
+#include <fcitx-utils/eventloopinterface.h>
 #include <fcitx-utils/key.h>
 #include <fcitx-utils/keysym.h>
-#include <fcitx-utils/keysymgen.h>
 #include <fcitx-utils/macros.h>
 #include <fcitx-utils/misc.h>
 #include <fcitx-utils/stringutils.h>
@@ -42,6 +42,7 @@
 #include <fstream>
 #include <functional>
 #include <future>
+#include <ios>
 #include <istream>
 #include <iterator>
 #include <libime/core/languagemodel.h>
@@ -52,6 +53,7 @@
 #include <memory>
 #include <optional>
 #include <ostream>
+#include <regex>
 #include <stdexcept>
 #include <string>
 #include <string_view>
@@ -97,6 +99,8 @@
 
 namespace fcitx {
 
+namespace {
+
 FCITX_DEFINE_LOG_CATEGORY(pinyin, "pinyin");
 
 #define PINYIN_DEBUG() FCITX_LOGC(pinyin, Debug)
@@ -143,50 +147,6 @@ predictCandidateList(PinyinEngine *engine, const std::vector<T> &words) {
     return candidateList;
 }
 
-PinyinState::PinyinState(PinyinEngine *engine) : context_(engine->ime()) {
-    context_.setMaxSentenceLength(35);
-}
-
-void PinyinEngine::initPredict(InputContext *inputContext) {
-    inputContext->inputPanel().reset();
-
-    auto *state = inputContext->propertyFor(&factory_);
-    auto &context = state->context_;
-    auto lmState = context.state();
-    state->predictWords_ = context.selectedWords();
-    auto words =
-        prediction_.predict(lmState, context.selectedWords(),
-                            context.selectedWordsWithPinyin().back().second,
-                            *config_.predictionSize);
-    if (auto candidateList = predictCandidateList(this, words)) {
-        auto &inputPanel = inputContext->inputPanel();
-        inputPanel.setCandidateList(std::move(candidateList));
-    } else {
-        state->predictWords_.reset();
-    }
-    inputContext->updatePreedit();
-    inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
-}
-
-void PinyinEngine::updatePredict(InputContext *inputContext) {
-    inputContext->inputPanel().reset();
-
-    auto *state = inputContext->propertyFor(&factory_);
-    assert(state->predictWords_.has_value());
-    auto words =
-        prediction_.predict(*state->predictWords_, *config_.predictionSize);
-    if (auto candidateList = predictCandidateList(this, words)) {
-        auto &inputPanel = inputContext->inputPanel();
-        inputPanel.setCandidateList(std::move(candidateList));
-    } else {
-        // Clear if we can't do predict.
-        // This help other code to detect whether we are in predict.
-        state->predictWords_.reset();
-    }
-    inputContext->updatePreedit();
-    inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
-}
-
 std::tuple<bool, int> englishNess(const std::string &input, bool sp) {
     const auto pys = stringutils::split(input, " ");
     constexpr int fullWeight = -2;
@@ -239,6 +199,52 @@ bool isStroke(const std::string &input) {
                        [](char c) { return py.count(c); });
 }
 
+} // namespace
+
+PinyinState::PinyinState(PinyinEngine *engine) : context_(engine->ime()) {
+    context_.setMaxSentenceLength(35);
+}
+
+void PinyinEngine::initPredict(InputContext *inputContext) {
+    inputContext->inputPanel().reset();
+
+    auto *state = inputContext->propertyFor(&factory_);
+    auto &context = state->context_;
+    auto lmState = context.state();
+    state->predictWords_ = context.selectedWords();
+    auto words =
+        prediction_.predict(lmState, context.selectedWords(),
+                            context.selectedWordsWithPinyin().back().second,
+                            *config_.predictionSize);
+    if (auto candidateList = predictCandidateList(this, words)) {
+        auto &inputPanel = inputContext->inputPanel();
+        inputPanel.setCandidateList(std::move(candidateList));
+    } else {
+        state->predictWords_.reset();
+    }
+    inputContext->updatePreedit();
+    inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
+}
+
+void PinyinEngine::updatePredict(InputContext *inputContext) {
+    inputContext->inputPanel().reset();
+
+    auto *state = inputContext->propertyFor(&factory_);
+    assert(state->predictWords_.has_value());
+    auto words =
+        prediction_.predict(*state->predictWords_, *config_.predictionSize);
+    if (auto candidateList = predictCandidateList(this, words)) {
+        auto &inputPanel = inputContext->inputPanel();
+        inputPanel.setCandidateList(std::move(candidateList));
+    } else {
+        // Clear if we can't do predict.
+        // This help other code to detect whether we are in predict.
+        state->predictWords_.reset();
+    }
+    inputContext->updatePreedit();
+    inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
+}
+
 #ifdef FCITX_HAS_LUA
 std::vector<std::string>
 PinyinEngine::luaCandidateTrigger(InputContext *ic,
@@ -535,9 +541,7 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
         if (pinyinhelper() && context.selectedLength() == 0 &&
             isStroke(context.userInput())) {
             int limit = (context.userInput().size() + 4) / 5;
-            if (limit > 3) {
-                limit = 3;
-            }
+            limit = std::min(limit, 3);
             auto results = pinyinhelper()->call<IPinyinHelper::lookupStroke>(
                 context.userInput(), limit);
 
@@ -576,9 +580,7 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
                     int actualPos = (lastPos >= extraCandidate->order())
                                         ? lastPos
                                         : extraCandidate->order();
-                    if (actualPos > candidateList->totalSize()) {
-                        actualPos = candidateList->totalSize();
-                    }
+                    actualPos = std::min(actualPos, candidateList->totalSize());
                     // Rewrap it into CandidateWord.
                     std::unique_ptr<CandidateWord> cand(
                         &extraCandidate.release()->toCandidateWord());
@@ -912,7 +914,7 @@ void PinyinEngine::loadExtraDict() {
             continue;
         }
         PINYIN_DEBUG() << "Loading extra dictionary: " << file.first;
-        loadDict(std::move(file.second), tasks_);
+        loadDict(file.second, tasks_);
     }
 }
 
@@ -983,52 +985,8 @@ void PinyinEngine::populateConfig() {
     ime_->setPreeditMode(*config_.showActualPinyinInPreedit
                              ? libime::PinyinPreeditMode::Pinyin
                              : libime::PinyinPreeditMode::RawText);
-    if (*config_.shuangpinProfile == ShuangpinProfileEnum::Custom) {
-        auto file = StandardPath::global().open(StandardPath::Type::PkgConfig,
-                                                "pinyin/sp.dat", O_RDONLY);
-        if (file.isValid()) {
-            try {
-                boost::iostreams::stream_buffer<
-                    boost::iostreams::file_descriptor_source>
-                    buffer(file.fd(), boost::iostreams::file_descriptor_flags::
-                                          never_close_handle);
-                std::istream in(&buffer);
-                ime_->setShuangpinProfile(
-                    std::make_shared<libime::ShuangpinProfile>(in));
-            } catch (const std::exception &e) {
-                PINYIN_ERROR() << e.what();
-            }
-        } else {
-            PINYIN_ERROR() << "Failed to open shuangpin profile.";
-        }
-    } else {
-        libime::ShuangpinBuiltinProfile profile;
-#define TRANS_SP_PROFILE(PROFILE)                                              \
-    case ShuangpinProfileEnum::PROFILE:                                        \
-        profile = libime::ShuangpinBuiltinProfile::PROFILE;                    \
-        break;
-        switch (*config_.shuangpinProfile) {
-            TRANS_SP_PROFILE(Ziranma)
-            TRANS_SP_PROFILE(MS)
-            TRANS_SP_PROFILE(Ziguang)
-            TRANS_SP_PROFILE(ABC)
-            TRANS_SP_PROFILE(Zhongwenzhixing)
-            TRANS_SP_PROFILE(PinyinJiajia)
-            TRANS_SP_PROFILE(Xiaohe)
-        default:
-            profile = libime::ShuangpinBuiltinProfile::Ziranma;
-            break;
-        }
-        ime_->setShuangpinProfile(
-            std::make_shared<libime::ShuangpinProfile>(profile));
-    }
-
-    // Always set a profile to avoid crash.
-    if (!ime_->shuangpinProfile()) {
-        ime_->setShuangpinProfile(std::make_shared<libime::ShuangpinProfile>(
-            libime::ShuangpinBuiltinProfile::Ziranma));
-    }
 
+    // Setup fuzzy flags.
     libime::PinyinFuzzyFlags flags;
     const auto &fuzzyConfig = *config_.fuzzyConfig;
 #define SET_FUZZY_FLAG(VAR, ENUM)                                              \
@@ -1055,50 +1013,86 @@ void PinyinEngine::populateConfig() {
     SET_FUZZY_FLAG(s, S_SH)
     SET_FUZZY_FLAG(z, Z_ZH)
 
-    std::shared_ptr<libime::PinyinCorrectionProfile> correctionProfile;
-    switch (*fuzzyConfig.correction) {
-    case CorrectionLayout::None:
-        break;
-    case CorrectionLayout::Qwerty:
-        correctionProfile = std::make_shared<libime::PinyinCorrectionProfile>(
-            libime::BuiltinPinyinCorrectionProfile::Qwerty);
-        break;
-    }
-
-    if (correctionProfile) {
-        flags |= libime::PinyinFuzzyFlag::Correction;
+    {
+        std::shared_ptr<libime::PinyinCorrectionProfile> correctionProfile;
+        switch (*fuzzyConfig.correction) {
+        case CorrectionLayout::None:
+            break;
+        case CorrectionLayout::Qwerty:
+            correctionProfile =
+                std::make_shared<libime::PinyinCorrectionProfile>(
+                    libime::BuiltinPinyinCorrectionProfile::Qwerty);
+            break;
+        }
         ime_->setCorrectionProfile(std::move(correctionProfile));
     }
 
+    if (ime_->correctionProfile()) {
+        flags |= libime::PinyinFuzzyFlag::Correction;
+    }
     ime_->setFuzzyFlags(flags);
 
-    quickphraseTriggerDict_.clear();
-    for (std::string_view prefix : *config_.quickphraseTrigger) {
-        if (prefix.empty()) {
-            continue;
+    if (*config_.shuangpinProfile == ShuangpinProfileEnum::Custom) {
+        auto file = StandardPath::global().open(StandardPath::Type::PkgConfig,
+                                                "pinyin/sp.dat", O_RDONLY);
+        if (file.isValid()) {
+            try {
+                boost::iostreams::stream_buffer<
+                    boost::iostreams::file_descriptor_source>
+                    buffer(file.fd(), boost::iostreams::file_descriptor_flags::
+                                          never_close_handle);
+                std::istream in(&buffer);
+                ime_->setShuangpinProfile(
+                    std::make_shared<libime::ShuangpinProfile>(
+                        in, ime_->correctionProfile().get()));
+            } catch (const std::exception &e) {
+                PINYIN_ERROR() << e.what();
+            }
+        } else {
+            PINYIN_ERROR() << "Failed to open shuangpin profile.";
         }
-        auto length = utf8::lengthValidated(prefix);
-        if (length == utf8::INVALID_LENGTH || length < 1) {
-            continue;
+    } else {
+        libime::ShuangpinBuiltinProfile profile;
+#define TRANS_SP_PROFILE(PROFILE)                                              \
+    case ShuangpinProfileEnum::PROFILE:                                        \
+        profile = libime::ShuangpinBuiltinProfile::PROFILE;                    \
+        break;
+        switch (*config_.shuangpinProfile) {
+            TRANS_SP_PROFILE(Ziranma)
+            TRANS_SP_PROFILE(MS)
+            TRANS_SP_PROFILE(Ziguang)
+            TRANS_SP_PROFILE(ABC)
+            TRANS_SP_PROFILE(Zhongwenzhixing)
+            TRANS_SP_PROFILE(PinyinJiajia)
+            TRANS_SP_PROFILE(Xiaohe)
+        default:
+            profile = libime::ShuangpinBuiltinProfile::Ziranma;
+            break;
         }
-        auto latinPartLength =
-            utf8::ncharByteLength(prefix.begin(), length - 1);
-        auto latinPart = prefix.substr(0, latinPartLength);
+        ime_->setShuangpinProfile(std::make_shared<libime::ShuangpinProfile>(
+            profile, ime_->correctionProfile().get()));
+    }
 
-        if (std::any_of(latinPart.begin(), latinPart.end(), [](char c) {
-                return !charutils::islower(c) && !charutils::isupper(c);
-            })) {
-            continue;
-        }
+    // Always set a profile to avoid crash.
+    if (!ime_->shuangpinProfile()) {
+        ime_->setShuangpinProfile(std::make_shared<libime::ShuangpinProfile>(
+            libime::ShuangpinBuiltinProfile::Ziranma,
+            ime_->correctionProfile().get()));
+    }
 
-        const uint32_t trigger =
-            utf8::getChar(prefix.begin() + latinPartLength, prefix.end());
-        if (trigger && trigger != utf8::INVALID_CHAR &&
-            trigger != utf8::NOT_ENOUGH_SPACE) {
-            quickphraseTriggerDict_[std::string(latinPart)].insert(trigger);
+    quickphraseTriggerRegex_.clear();
+    for (const std::string &regStr : *config_.quickphraseTriggerRegex) {
+        try {
+            std::regex reg{regStr};
+            quickphraseTriggerRegex_.push_back(std::move(reg));
+        } catch (const std::exception &e) {
+            PINYIN_DEBUG() << "Invalid regular expression: \"" << regStr
+                           << "\", " << e.what();
         }
     }
-    PINYIN_DEBUG() << "Quick Phrase Trigger Dict: " << quickphraseTriggerDict_;
+    PINYIN_DEBUG() << "Quick Phrase Trigger Regex size: "
+                   << quickphraseTriggerRegex_.size();
+
     ime_->dict()->setFlags(libime::TrieDictionary::UserDict + 1,
                            *config_.chaiziEnabled
                                ? libime::PinyinDictFlag::FullMatch
@@ -1292,10 +1286,13 @@ bool PinyinEngine::handle2nd3rdSelection(KeyEvent &event) {
             if (keyReleased == idx &&
                 keyReleasedIndex == event.key().keyListIndex(keyHandler.list)) {
                 if (isModifier) {
-                    if (keyHandler.selection < candidateList->size()) {
+                    if (instance_->globalConfig().checkModifierOnlyKeyTimeout(
+                            state->lastKeyPressedTime_) &&
+                        keyHandler.selection < candidateList->size()) {
                         candidateList->candidate(keyHandler.selection)
                             .select(inputContext);
                     }
+                    state->lastKeyPressedTime_ = 0;
                     event.filterAndAccept();
                     return true;
                 }
@@ -1314,6 +1311,7 @@ bool PinyinEngine::handle2nd3rdSelection(KeyEvent &event) {
                 state->keyReleased_ = idx;
                 state->keyReleasedIndex_ = keyIdx;
                 if (isModifier) {
+                    state->lastKeyPressedTime_ = now(CLOCK_MONOTONIC);
                     // don't forward to input method, but make it pass
                     // through to client.
                     event.filter();
@@ -1332,7 +1330,8 @@ bool PinyinEngine::handle2nd3rdSelection(KeyEvent &event) {
     return false;
 }
 
-bool PinyinEngine::handleCandidateList(KeyEvent &event) {
+bool PinyinEngine::handleCandidateList(
+    KeyEvent &event, const std::shared_future<uint32_t> &keyChr) {
     auto *inputContext = event.inputContext();
     auto candidateList = inputContext->inputPanel().candidateList();
     if (!candidateList) {
@@ -1343,10 +1342,7 @@ bool PinyinEngine::handleCandidateList(KeyEvent &event) {
         !state->predictWords_) {
         if (!candidateList->empty()) {
             event.filterAndAccept();
-            int idx = candidateList->cursorIndex();
-            if (idx < 0) {
-                idx = 0;
-            }
+            int idx = std::max(candidateList->cursorIndex(), 0);
             candidateList->candidate(idx).select(inputContext);
             return true;
         }
@@ -1376,7 +1372,7 @@ bool PinyinEngine::handleCandidateList(KeyEvent &event) {
                 return true;
             }
             // Only let key go through if it can reach handlePunc.
-            auto c = Key::keySymToUnicode(event.key().sym());
+            auto c = keyChr.get();
             if (event.key().hasModifier() || !c) {
                 event.filterAndAccept();
                 return true;
@@ -1464,8 +1460,18 @@ void PinyinEngine::updateStroke(InputContext *inputContext) {
                 }
             }
             if (strokeMatched) {
-                candidateList->append<StrokeFilterCandidateWord>(
-                    this, inputContext, candidate.text(), i);
+                // PinyinCandidateWord is the only forgettable.
+                if (dynamic_cast<const PinyinCandidateWord *>(&candidate)) {
+                    candidateList->append<StrokeFilterCandidateWord<
+                        FilteredForgettableCandidate,
+                        FilteredInsertableAsCustomPhrase>>(this, inputContext,
+                                                           candidate.text(), i);
+                } else if (dynamic_cast<const InsertableAsCustomPhraseInterface
+                                            *>(&candidate)) {
+                    candidateList->append<StrokeFilterCandidateWord<
+                        FilteredInsertableAsCustomPhrase>>(this, inputContext,
+                                                           candidate.text(), i);
+                }
             }
         }
     }
@@ -1541,6 +1547,8 @@ void PinyinEngine::resetForgetCandidate(InputContext *inputContext) const {
 void PinyinEngine::forgetCandidate(InputContext *inputContext, size_t index) {
     auto *state = inputContext->propertyFor(&factory_);
 
+    const std::string currentInput = state->context_.userInput();
+
     if (index < state->context_.candidatesToCursor().size()) {
         const auto &sentence = state->context_.candidatesToCursor()[index];
         // If this is a word, remove it from user dict.
@@ -1555,6 +1563,9 @@ void PinyinEngine::forgetCandidate(InputContext *inputContext, size_t index) {
     }
     resetForgetCandidate(inputContext);
     doReset(inputContext);
+
+    state->context_.type(currentInput);
+    updateUI(inputContext);
 }
 
 void PinyinEngine::saveCustomPhrase() {
@@ -1612,7 +1623,8 @@ void PinyinEngine::deleteCustomPhrase(InputContext *inputContext,
     saveCustomPhrase();
 }
 
-bool PinyinEngine::handleStrokeFilter(KeyEvent &event) {
+bool PinyinEngine::handleStrokeFilter(
+    KeyEvent &event, const std::shared_future<uint32_t> &keyChr) {
     auto *inputContext = event.inputContext();
     auto candidateList = inputContext->inputPanel().candidateList();
     auto *state = inputContext->propertyFor(&factory_);
@@ -1650,7 +1662,7 @@ bool PinyinEngine::handleStrokeFilter(KeyEvent &event) {
         }
     }
 
-    if (handleCandidateList(event)) {
+    if (handleCandidateList(event, keyChr)) {
         return true;
     }
     // Skip all key combination.
@@ -1676,7 +1688,7 @@ bool PinyinEngine::handleStrokeFilter(KeyEvent &event) {
         return true;
     }
     // if it gonna commit something
-    auto c = Key::keySymToUnicode(event.key().sym());
+    auto c = keyChr.get();
     if (!c) {
         return true;
     }
@@ -1736,7 +1748,8 @@ bool PinyinEngine::handleForgetCandidate(KeyEvent &event) {
     return true;
 }
 
-bool PinyinEngine::handlePunc(KeyEvent &event) {
+bool PinyinEngine::handlePunc(KeyEvent &event,
+                              const std::shared_future<uint32_t> &keyChr) {
     auto *inputContext = event.inputContext();
     auto candidateList = inputContext->inputPanel().candidateList();
     auto *state = inputContext->propertyFor(&factory_);
@@ -1744,7 +1757,7 @@ bool PinyinEngine::handlePunc(KeyEvent &event) {
         return false;
     }
     // if it gonna commit something
-    auto c = Key::keySymToUnicode(event.key().sym());
+    auto c = keyChr.get();
     if (event.key().hasModifier() || !c) {
         // Handle quick phrase with modifier
         if (event.key().check(*config_.quickphraseKey) && quickphrase()) {
@@ -1759,6 +1772,18 @@ bool PinyinEngine::handlePunc(KeyEvent &event) {
         candidateList->candidate(0).select(inputContext);
     }
 
+    // We can't emit punctuation right now, since it's not yet fully selected.
+    // There is an option to split the current input, however, it may not be
+    // sane to do so based on the partial input and cursor location. E.g.
+    // current input is n|i hao, and you press ",". Candidate will be filtered
+    // until "ni", instead of "n". which means you are not inserting "," between
+    // n & i. Thus, let's just make it select the default candidate.
+    // Also, it can not work with punctuation candidate, or type punc in pair.
+    if (!state->context_.empty()) {
+        event.filterAndAccept();
+        return true;
+    }
+
     std::string punc;
     std::string puncAfter;
     // skip key pad
@@ -1878,6 +1903,15 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     auto *inputContext = event.inputContext();
     auto *state = inputContext->propertyFor(&factory_);
 
+    std::shared_future<uint32_t> keyChr =
+        std::async(std::launch::deferred, [sym = event.key().sym()] {
+            return Key::keySymToUnicode(sym);
+        }).share();
+    std::shared_future<std::string> keyStr =
+        std::async(std::launch::deferred, [&keyChr] {
+            return utf8::UCS4ToUTF8(keyChr.get());
+        }).share();
+
     // 2nd/3rd selection is allowed to be modifier only, handle them before we
     // skip the release.
     if (handle2nd3rdSelection(event)) {
@@ -1906,7 +1940,7 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     bool lastIsPunc = state->lastIsPunc_;
     state->lastIsPunc_ = false;
 
-    if (handleStrokeFilter(event)) {
+    if (handleStrokeFilter(event, keyChr)) {
         return;
     }
 
@@ -1915,7 +1949,7 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     }
 
     // handle number key selection and prev/next page/candidate.
-    if (handleCandidateList(event)) {
+    if (handleCandidateList(event, keyChr)) {
         return;
     }
 
@@ -1936,7 +1970,7 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
         }
     }
 
-    auto checkSp = [this](const KeyEvent &event, PinyinState *state) {
+    auto checkSp = [this, &keyChr](const KeyEvent &event, PinyinState *state) {
         auto shuangpinProfile = ime_->shuangpinProfile();
         if (!state->context_.useShuangpin() || !shuangpinProfile ||
             !event.key().isSimple()) {
@@ -1944,26 +1978,40 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
         }
         // event.key().isSimple() make sure the return value is within range of
         // char.
-        char chr = Key::keySymToUnicode(event.key().sym());
+        char chr = keyChr.get();
         return (!state->context_.empty() &&
                 shuangpinProfile->validInput().count(chr)) ||
                (state->context_.empty() &&
                 shuangpinProfile->validInitial().count(chr));
     };
 
-    if (!event.key().hasModifier() && quickphrase()) {
-        const auto iter =
-            quickphraseTriggerDict_.find(state->context_.userInput());
-        if (iter != quickphraseTriggerDict_.end() && !iter->second.empty() &&
-            iter->second.count(Key::keySymToUnicode(event.key().sym()))) {
-            std::string text = state->context_.userInput();
-            text.append(Key::keySymToUTF8(event.key().sym()));
-            doReset(inputContext);
-            quickphrase()->call<IQuickPhrase::trigger>(inputContext, "", "", "",
-                                                       "", Key());
-            quickphrase()->call<IQuickPhrase::setBuffer>(inputContext, text);
-
-            return event.filterAndAccept();
+    if (!event.key().hasModifier() && quickphrase() &&
+        !quickphraseTriggerRegex_.empty() && !keyStr.get().empty() &&
+        state->context_.selectedLength() == 0 &&
+        state->context_.cursor() == state->context_.size()) {
+        std::string text =
+            stringutils::concat(state->context_.userInput(), keyStr.get());
+        for (const auto &reg : quickphraseTriggerRegex_) {
+            if (std::regex_search(text, reg,
+                                  std::regex_constants::match_default)) {
+                // Keep the current state before reset.
+                const std::string origin = state->context_.userInput();
+                doReset(inputContext);
+                quickphrase()->call<IQuickPhrase::trigger>(inputContext, "", "",
+                                                           "", "", Key());
+                quickphrase()->call<IQuickPhrase::setBufferWithRestoreCallback>(
+                    inputContext, text, origin,
+                    [this](InputContext *ic, const std::string &origin) {
+                        if (this->instance()->inputMethodEngine(ic) == this) {
+                            auto *state = ic->propertyFor(&factory_);
+                            doReset(ic);
+                            state->context_.type(origin);
+                            updateUI(ic);
+                        }
+                    });
+                event.filterAndAccept();
+                return;
+            }
         }
     }
 
@@ -1981,7 +2029,7 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
             return;
         }
         event.filterAndAccept();
-        if (!state->context_.type(Key::keySymToUTF8(event.key().sym()))) {
+        if (!state->context_.type(keyStr.get())) {
             return;
         }
     } else if (!state->context_.empty()) {
@@ -2125,7 +2173,7 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
             }
         }
     }
-    if (handlePunc(event)) {
+    if (handlePunc(event, keyChr)) {
         return;
     }
 
@@ -2239,7 +2287,8 @@ void PinyinEngine::invokeActionImpl(const InputMethodEntry &entry,
     if (event.cursor() < 0 ||
         event.action() != InvokeActionEvent::Action::LeftClick ||
         !inputContext->capabilityFlags().test(CapabilityFlag::Preedit)) {
-        return InputMethodEngineV3::invokeActionImpl(entry, event);
+        InputMethodEngineV3::invokeActionImpl(entry, event);
+        return;
     }
     auto [clientPreedit, _] = this->preedit(inputContext);
 
@@ -2248,7 +2297,8 @@ void PinyinEngine::invokeActionImpl(const InputMethodEntry &entry,
     if (inputPanel.clientPreedit().toString() != clientPreedit.toString() ||
         inputPanel.clientPreedit().cursor() != clientPreedit.cursor() ||
         cursor > utf8::length(preedit)) {
-        return InputMethodEngineV3::invokeActionImpl(entry, event);
+        InputMethodEngineV3::invokeActionImpl(entry, event);
+        return;
     }
 
     event.filter();
@@ -2412,5 +2462,3 @@ void PinyinEngine::cloudPinyinSelected(InputContext *inputContext,
 }
 
 } // namespace fcitx
-
-FCITX_ADDON_FACTORY(fcitx::PinyinEngineFactory)
