@@ -18,7 +18,7 @@
 #include <fcitx-utils/fs.h>
 #include <fcitx-utils/i18n.h>
 #include <fcitx-utils/key.h>
-#include <fcitx-utils/keysymgen.h>
+#include <fcitx-utils/keysym.h>
 #include <fcitx-utils/log.h>
 #include <fcitx-utils/macros.h>
 #include <fcitx-utils/misc.h>
@@ -576,4 +576,4 @@ void Punctuation::setSubConfig(const std::string &path,
     iter->second.save(lang);
 }
 
-FCITX_ADDON_FACTORY(PunctuationFactory);
+FCITX_ADDON_FACTORY_V2(punctuation, PunctuationFactory);
